/*****************************************
   name: GPS Clock Plus  Arduino Every
   
   function: Get GPS data and Write to 20x4 display
   This displays gps time and date, local time and grid square
   also lat, long and gps quality.  

   GPS Baud Rate 9600 using Serial1
   Serial Baud Rate 115200
   
   Use potentiometer to input time zone offset
   using a "uBlox Neo 6M" GPS module
 ********************************/
#define Debug false //general debugging
#define Debug_time false
#define Debug_location false
#define Debug_elevation false
#define Debug_satellites false
#define Debug_course false
#define Debug_speed false

#include <TimeLib.h>

#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x27, 20, 4); // set the LCD address to 0x27 for a 20 chars and 4 line display
// pins  +5v, gnd, 23 sda, 24 scl  

//Include Library code for GPS
#include <TinyGPS++.h>
TinyGPSPlus gps; //using Serial1 pins 1Tx 2Rx on Nano Every
//setup to do special receives
//TinyGPSCustom Act(gps, "GPGSA", 1); // $GPGSA sentence, 1st element  Active or Not
//TinyGPSCustom Mode(gps, "GPGSA", 2); // $GPGSA sentence, 2nd element  2D or 3D

//setup to get timezone data from potentiometer analog input
int TimeZonePin = 14;    // select the input pin for the potentiometer A0=pin 14
int TimeZoneValue = 0;  // variable to store the value coming from the sensor

unsigned long LLtime;
unsigned long LLcurrent;
boolean latlong = true;

// variable definitions
char Time[]     = "00:00:00 DD";
char utcTime[]  = "00:00:00 DD";
byte utcwday;
char Date[]     = "00/00/00";
char utcDate[]  = "00/00/00";
String dname[8] = { "Su", "Mo", "Tu","We","Th","Fr","Sa"};
String wday = "  ";

byte last_second, Second, Minute, Hour, last_minute, Day, Month;
int Year;

void setup() {
  // Open serial communications and wait for port to open:
  Serial.begin(115200);
  // set the data rate for the SoftwareSerial port
  Serial1.begin(9600);
  Serial.println("Starting the GPS hunt ... ");
  lcd.init();  //initialize the lcd
  lcd.backlight();  //open the backlight
  lcd.setCursor ( 0, 0 );            // go to the top left corner
  lcd.print("      GPS Clock  "); // write this string on the top row
  lcd.setCursor ( 0, 1 );            // go to the 2nd row
  lcd.print("    Plus Grid Sq  "); // pad string with spaces for centering
  lcd.setCursor ( 0, 2 );            // go to the third row
  lcd.print("       KG6JAZ       "); // pad with spaces for centering
  lcd.setCursor ( 0, 3 );            // go to the fourth row
  lcd.print("    Rod Ledbetter ");
  lcd.setCursor ( 0 , 1); 
  delay(6000);
  lcd.clear();   
  lcd.setCursor ( 0, 0 );            // go to the top left corner
  lcd.print("UTC Date Time Day");    // write this string on the top row
  lcd.setCursor ( 0, 1 );            // go to the 2nd row
  lcd.print("Lat Lon Grid Sq  ");    // pad string with spaces for centering
  lcd.setCursor ( 0, 2 );            // go to the third row
  lcd.print("Local Date Time Day "); // pad with spaces for centering
  lcd.setCursor ( 0, 3 );            // go to the fourth row
  lcd.print("Spd #Sats  Elev   ");
  lcd.setCursor ( 0 , 1); 
  delay(6000);
  lcd.clear(); 
}

void loop() {
  float flat, flng;
  //Get Time Zone Offset Hours can be used to do Daylight Savings Time Offset Also
  TimeZoneValue = analogRead(TimeZonePin);
  TimeZoneValue = map(TimeZoneValue,0,1023,-12,12); //12 hours before and after GMT
  LLcurrent = millis();  //get current timer info
  if (LLcurrent - LLtime >=4000) {
    latlong = !latlong;
    LLtime = LLcurrent;
  }
  while (Serial1.available() > 0) {
    if (gps.encode(Serial1.read())){
      if (gps.time.isUpdated()&&gps.date.isUpdated()){  //Get Time and Date - Utc and Local
        // get time from GPS module
        Minute = gps.time.minute();
        Second = gps.time.second();
        Hour   = gps.time.hour();
        Day   = gps.date.day();
        Month = gps.date.month();
        Year  = gps.date.year();
#if (Debug_time)
        Serial.print(F("TIME       Fix Age="));
        Serial.print(gps.time.age());
        Serial.print(F("ms Raw="));
        Serial.print(gps.time.value());
        Serial.print(F(" Hour="));
        Serial.print(gps.time.hour());
        Serial.print(F(" Minute="));
        Serial.print(gps.time.minute());
        Serial.print(F(" Second="));
        Serial.print(gps.time.second());
        Serial.print(F(" Hundredths="));
        Serial.println(gps.time.centisecond()); 
        Serial.print(F("DATE       Fix Age="));
        Serial.print(gps.date.age());
        Serial.print(F("ms Raw="));
        Serial.print(gps.date.value());
        Serial.print(F(" Year="));
        Serial.print(gps.date.year());
        Serial.print(F(" Month="));
        Serial.print(gps.date.month());
        Serial.print(F(" Day="));
        Serial.println(gps.date.day());
#endif           
        if(last_second != gps.time.second()) {  // if time has changed
          last_second = gps.time.second();
          // set current UTC time
          setTime(Hour, Minute, Second, Day, Month, Year);          
          // update utc time array
          utcTime[6] = second() / 10 + '0';
          utcTime[7] = second() % 10 + '0';
          utcTime[3]  = minute() / 10 + '0';
          utcTime[4] = minute() % 10 + '0';
          utcTime[0]  = hour()   / 10 + '0';
          utcTime[1]  = hour()   % 10 + '0';
          wday = dname[weekday()-1]; //add dow to display
          utcTime[9]  =  wday[0];        
          utcTime[10] =  wday[1]; 
          // update utc date array
          utcDate[6] = (year()  / 10) % 10 + '0';
          utcDate[7] =  year()  % 10 + '0';
          utcDate[0] =  month() / 10 + '0';
          utcDate[1] =  month() % 10 + '0';
          utcDate[3] =  day()   / 10 + '0';
          utcDate[4] =  day()   % 10 + '0';
          // add the offset to get local time       
          adjustTime(TimeZoneValue * 3600); //correct to local time
          // update local time array
          Time[6] = second() / 10 + '0';
          Time[7] = second() % 10 + '0';
          Time[3] = minute() / 10 + '0';
          Time[4] = minute() % 10 + '0';
          Time[0] = hour()   / 10 + '0';
          Time[1] = hour()   % 10 + '0';
          wday="";
          wday = dname[weekday()-1];
          Time[9]  =  wday[0];        
          Time[10] =  wday[1]; 
          // update local date array
          Date[6] = (year()  / 10) % 10 + '0';
          Date[7] =  year()  % 10 + '0';
          Date[0]  =  month() / 10 + '0';
          Date[1] =  month() % 10 + '0';
          Date[3]  =  day()   / 10 + '0';
          Date[4]  =  day()   % 10 + '0';
          // print time and Date
          //lcd.print(utcDate);      // print date (DD-MM-YYYY)
          lcd.setCursor(9, 0);       // move cursor to column 0 row 2
          lcd.print(utcTime);        // print time (HH:MM:SS)
          //lcd.setCursor(0, 2);     // move cursor to column 0 row 3
          //lcd.print(Date);         // print date (DD-MM-YYYY)
          lcd.setCursor(9, 2);       // move cursor to column 0 row 2
          lcd.print(Time);           // print time (HH:MM:SS)  
          // print date
          lcd.setCursor(0, 0);     // move cursor to column 0 row 3
          lcd.print(utcDate);      // print utc date (DD-MM-YYYY)
          lcd.setCursor(0, 2);     // move cursor to column 0 row 3
          lcd.print(Date);         // print local date (DD-MM-YYYY)           
        } //if second changed update
      } //time and date
      else if (gps.location.isUpdated()) { //Get Location
        #if (Debug_location)
        Serial.print(F("LOCATION   Fix Age="));
        Serial.print(gps.location.age());
        Serial.print(F("ms Raw Lat="));
        Serial.print(gps.location.rawLat().negative ? "-" : "+");
        Serial.print(gps.location.rawLat().deg);
        Serial.print("[+");
        Serial.print(gps.location.rawLat().billionths);
        Serial.print(F(" billionths],  Raw Long="));
        Serial.print(gps.location.rawLng().negative ? "-" : "+");
        Serial.print(gps.location.rawLng().deg);
        Serial.print("[+");
        Serial.print(gps.location.rawLng().billionths);
        Serial.print(F(" billionths],  Lat="));
        Serial.print(gps.location.lat(), 6);
        Serial.print(F(" Long="));
        Serial.println(gps.location.lng(), 6);
        #endif
        //Send to LCD
        flat=gps.location.lat();
        flng=gps.location.lng();
        lcd.setCursor(0 , 1);
        if (latlong) {
          lcd.print("Lt");
          printFloat(flat, 6); lcd.print("  ");  // printFloat(flng, 2);
        }
        else {
          lcd.print("Ln");
          printFloat(flng, 6); lcd.print("  "); 
        }
        lcd.setCursor(14 , 1);
        lcd.print(LatLonToGridSquare(flat, flng));
        //lcd.print(":"+LatLonToGridSquare(flat, flng));
       }  //Lat Long    
       else if (gps.speed.isUpdated()) { // get speed in MPH
         #if (Debug_speed)
         Serial.print(F("SPEED      Fix Age="));
         Serial.print(gps.speed.age());
         Serial.print(F("ms Raw="));
         Serial.print(gps.speed.value());
         Serial.print(F(" Knots="));
         Serial.print(gps.speed.knots());
         Serial.print(F(" MPH="));
         Serial.print(gps.speed.mph());
         Serial.print(F(" m/s="));
         Serial.print(gps.speed.mps());
         Serial.print(F(" km/h="));
         Serial.println(gps.speed.kmph());
         #endif
         lcd.setCursor (0,3);
         lcd.print("S     ");
         lcd.setCursor (0,3);
         lcd.print("S ");
         //lcd.print(gps.speed.mph()); 
         printFloat(gps.speed.mph(), 0);
       }//Speed
       else if (gps.altitude.isUpdated()) {//get altitude in feet
         #if (Debug_altitude)
         Serial.print(F("ALTITUDE   Fix Age="));
         Serial.print(gps.altitude.age());
         Serial.print(F("ms Raw="));
         Serial.print(gps.altitude.value());
         Serial.print(F(" Meters="));
         Serial.print(gps.altitude.meters());
         Serial.print(F(" Miles="));
         Serial.print(gps.altitude.miles());
         Serial.print(F(" KM="));
         Serial.print(gps.altitude.kilometers());
         Serial.print(F(" Feet="));
         Serial.println(gps.altitude.feet());
         #endif
         lcd.setCursor (13,3);
         lcd.print(F("E ")); 
         lcd.print(int(gps.altitude.feet())); 
         lcd.print(F("  ")); 
       } //Altitude
       else if (gps.satellites.isUpdated()){
         #if (Debug_satellites)
         Serial.print(F("SATELLITES Fix Age="));
         Serial.print(gps.satellites.age());
         Serial.print(F("ms Value="));
         Serial.println(gps.satellites.value());
         #endif
         lcd.setCursor (8,3);     
         lcd.print(F("S "));
         if (gps.satellites.value()>20){
           lcd.print("?");       
         }
         else
           lcd.print(gps.satellites.value());
         lcd.print(F(" ")); 
       } //Satellites
       /*
       else if (gps.course.isUpdated()){
         #if (Debug_course)
         Serial.print(F("COURSE     Fix Age="));
         Serial.print(gps.course.age());
         Serial.print(F("ms Raw="));
         Serial.print(gps.course.value());
         Serial.print(F(" Deg="));
         Serial.println(gps.course.deg());
         #endif
         lcd.setCursor (6,3);     
         lcd.print(F("C    "));
         lcd.setCursor (8,3); 
         lcd.print(int(gps.course.deg()));
//         lcd.print(F(" ")); 
       } //Course
       */
#if (Debug)     
       Serial.print(F("DIAGS      Chars="));
       Serial.print(gps.charsProcessed());
       Serial.print(F(" Sentences-with-Fix="));
       Serial.print(gps.sentencesWithFix());
       Serial.print(F(" Failed-checksum="));
       Serial.print(gps.failedChecksum());
       Serial.print(F(" Passed-checksum="));
       Serial.println(gps.passedChecksum());
#endif
       if (gps.charsProcessed() < 10)
         Serial.println(F("WARNING: No GPS data.  Check wiring."));
     } //end if encode       
  } //end while
} //end of Loop

void printFloat(double number, int digits){
  // Handle negative numbers
  if (number < 0.0) {
     lcd.print('-');     
     number = -number;
  }
  // Round correctly so that print(1.999, 2) prints as "2.00"
  double rounding = 0.5;
  for (uint8_t i=0; i<digits; ++i)
    rounding /= 10.0;
  number += rounding;
  // Extract the integer part of the number and print it
  unsigned long int_part = (unsigned long)number;
  double remainder = number - (double)int_part;
  lcd.print(int_part);
  // Print the decimal point, but only if there are digits beyond
  if (digits > 0)
    lcd.print("."); 
  // Extract digits from the remainder one at a time
  while (digits-- > 0) {
    remainder *= 10.0;
    int toPrint = int(remainder);
    lcd.print(toPrint);    
    remainder -= toPrint;
  }
}

static String LatLonToGridSquare(float lat, float lon){
    float adjLat, adjLon;
    String UPLat, UPLon;
    String nLat, nLon;
    String LOLat, LOLon;
    float rLat, rLon;
    String U = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    String L = "abcdefghijklmnopqrstuvwxyz";

    if (!(-180<=lon<180))
      {
      lcd.print("Err");
      return;
      }
    if (!(-90<=lat<90))
      {
      lcd.print("Err");
      return;
      }
    adjLat = lat + 90;
    adjLon = lon + 180;
    UPLat = U.charAt(int(adjLat / 10));
    UPLon = U.charAt(int(adjLon / 20));
    nLat = String(int(adjLat) % 10);
    nLon = String(int(adjLon / 2) % 10);
    rLat = (adjLat - int(adjLat)) * 60;
    rLon = (adjLon - 2 * int(adjLon / 2)) * 60;
    LOLat = L[int(rLat / 2.5)];
    LOLon = L[int(rLon / 5.0)];
    String locator = "" + UPLon + UPLat + nLon + nLat + LOLon + LOLat;
    return locator;
}